(* Abstract Syntax Tree and functions for printing it *)

type op = Add | Sub | Mult | Div | Equal | Neq | Less | Leq | Greater | Geq |
          And | Or

type uop = Neg | Not | Inc | Dec

(* type prm_typ = Int | Bool | Float | Void | String *)
type high_typ = LIST | DICT | SET

type typ = 
    Int | Bool | Float | Void | String | Func
  | CompositeType of high_typ * typ list
  | FunctionType of typ list


and func_decl = {
    typ : typ;
    fname : string;
    formals : bind list;
    body : stmt list;
  }

and bind = typ * string

and expr =
    Literal of int
  | Fliteral of string
  | BoolLit of bool
  | StrLit of string
  | FunctionLit of func_decl
  | GetSlice of string * expr * expr
  | GetIndex of string * expr
  | Id of string
  | Binop of expr * op * expr
  | Unop of uop * expr
  | IndexAssign of string * expr * expr
  | Assign of string * expr
  | ListExpr of expr list
  | Call of string * expr list
  | IndexUnop of string * expr * uop
  | Noexpr

and stmt =
    Block of stmt list
  | Expr of expr
  | Return of expr
  | If of expr * stmt * stmt
  | For of expr * expr * expr * stmt
  | Forin of string * expr * stmt
  | While of expr * stmt
  | VarDecl of bind
  | VarDeclAssign of bind * expr

type program = bind list * func_decl list



(* Pretty-printing functions *)

let string_of_op = function
    Add -> "+"
  | Sub -> "-"
  | Mult -> "*"
  | Div -> "/"
  | Equal -> "=="
  | Neq -> "!="
  | Less -> "<"
  | Leq -> "<="
  | Greater -> ">"
  | Geq -> ">="
  | And -> "&&"
  | Or -> "||"

let string_of_uop = function
    Neg -> "-"
  | Not -> "!"
  | Inc -> "++"
	| Dec -> "--"

let string_of_high_typ = function
    LIST -> "List"
  | DICT -> "Dict"
  | SET ->  "Set"

let rec string_of_typ = function
    Int -> "int"
  | Bool -> "bool"
  | Float -> "float"
  | Void -> "void"
  | String -> "string"
  | Func -> "func"
  | CompositeType(ht, tl) -> 
      string_of_high_typ ht ^ "<" ^ String.concat ", " (List.map string_of_typ tl) ^ ">"
  | FunctionType(rt::rest) ->
      string_of_typ rt ^ ":" ^ String.concat ", " (List.map string_of_typ rest) ^ ""

let string_of_vdecl (t, id) = string_of_typ t ^ " " ^ id ^ ";\n"

let string_of_formal (t, id) = string_of_typ t ^ " " ^ id

let rec string_of_expr = function
    Literal(l) -> string_of_int l
  | Fliteral(l) -> l
  | BoolLit(true) -> "true"
  | BoolLit(false) -> "false"
  | StrLit(s) -> "\"" ^ s ^ "\""
  | FunctionLit(l) -> string_of_typ l.typ ^ " " ^
    l.fname ^ "(" ^ String.concat ", " (List.map string_of_formal l.formals) ^
    ")\n{}\n" 
  | Id(s) -> s
  | IndexAssign(v, e1, e2) -> v ^ "[" ^ string_of_expr e1 ^ "] = " ^ string_of_expr e2
  | GetIndex(v, e2) -> v ^ "[" ^ string_of_expr e2 ^ "]"
  | GetSlice(v, e1, e2) -> v ^ "[" ^ string_of_expr e1 ^ ":" ^ string_of_expr e2 ^ "]"
  | Binop(e1, o, e2) ->
      string_of_expr e1 ^ " " ^ string_of_op o ^ " " ^ string_of_expr e2
  | Unop(o, e) -> 
    (match o with 
    | Neg | Not -> string_of_uop o ^ string_of_expr e
    | Inc | Dec -> string_of_expr e ^ string_of_uop o
    )
  | ListExpr(el) -> "[ " ^ String.concat ", " (List.map string_of_expr el) ^ " ]"
  | Assign(v, e) -> v ^ " = " ^ string_of_expr e
  | Call(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
  | IndexUnop(v, e, o) ->
      v ^ "[" ^ string_of_expr e ^ "]" ^ string_of_uop o
  | Noexpr -> ""

let rec string_of_stmt = function
    Block(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | Expr(expr) -> string_of_expr expr ^ ";\n"
  | Return(expr) -> "return " ^ string_of_expr expr ^ ";\n"
  | If(e, s, Block([])) -> "if (" ^ string_of_expr e ^ ")\n" ^ string_of_stmt s
  | If(e, s1, s2) ->  "if (" ^ string_of_expr e ^ ")\n" ^
      string_of_stmt s1 ^ "else\n" ^ string_of_stmt s2
  | For(e1, e2, e3, s) ->
      "for (" ^ string_of_expr e1  ^ " ; " ^ string_of_expr e2 ^ " ; " ^
      string_of_expr e3  ^ ") " ^ string_of_stmt s
  | Forin(id1, id2, s) ->
      "for " ^ id1 ^ " in " ^ string_of_expr id2 ^ " " ^ string_of_stmt s
  | While(e, s) -> "while (" ^ string_of_expr e ^ ") " ^ string_of_stmt s
  | VarDecl(t, id) -> string_of_typ t ^ " " ^ id ^ ";\n" 
  | VarDeclAssign((t, id), e) -> string_of_typ t ^ " " ^ id ^ " = " ^ string_of_expr e ^ ";\n"



let string_of_fdecl fdecl =
  string_of_typ fdecl.typ ^ " " ^
  fdecl.fname ^ "(" ^ String.concat ", " (List.map string_of_formal fdecl.formals) ^
  ")\n{\n" ^
  String.concat "" (List.map string_of_stmt fdecl.body) ^
  "}\n"

let string_of_program (vars, funcs) =
  String.concat "" (List.map string_of_vdecl vars) ^ "\n" ^
  String.concat "\n" (List.map string_of_fdecl funcs)
