(* Code generation: translate takes a semantically checked AST and
produces LLVM IR

LLVM tutorial: Make sure to read the OCaml version of the tutorial

http://llvm.org/docs/tutorial/index.html

Detailed documentation on the OCaml LLVM library:

http://llvm.moe/
http://llvm.moe/ocaml/

*)

module L = Llvm
module A = Ast
open Sast
module StringMap = Map.Make (String)

(* translate : Sast.program -> Llvm.module *)
let translate ((globals, functions), top_funcs) =
  let context = L.global_context () in

  (* Create the LLVM compilation module into which
     we will generate code *)
  let the_module = L.create_module context "FFBB" in

  (* Get types from the context *)
  let i32_t = L.i32_type context
  and i8_t = L.i8_type context
  and i1_t = L.i1_type context
  and float_t = L.double_type context
  and str_t = L.pointer_type (L.i8_type context)
  and void_t = L.void_type context in
  let list_t typ ltyp =
    let typ_str = A.string_of_typ (A.CompositeType (A.LIST, [ typ ])) in
    let t = L.named_struct_type context typ_str in
    let list_body =
      [| i32_t; i32_t; i32_t; L.pointer_type ltyp |]
      (* length, type, capacity, pointer *)
    in
    ignore (L.struct_set_body t list_body false);
    t
  in
  let list_int_t = list_t A.Int i32_t
  and list_bool_t = list_t A.Bool i1_t
  and list_float_t = list_t A.Float float_t in
  let set_t =
    let t = L.named_struct_type context "TreeSet" in
    let set_body = [| i32_t; i32_t; L.pointer_type i8_t |] in
    ignore (L.struct_set_body t set_body false);
    t
  in
  let dict_t =
    let t = L.named_struct_type context "TreeDict" in
    let set_body = [| i32_t; i32_t; L.pointer_type i8_t |] in
    ignore (L.struct_set_body t set_body false);
    t
  in
  (* Return the LLVM type for a FFBB type *)
  let rec ltype_of_typ = function
    | A.Int -> i32_t
    | A.Bool -> i1_t
    | A.Float -> float_t
    | A.Void -> void_t
    | A.String -> str_t
    | A.FunctionType (rt :: rest) ->
        let llargs =
          List.fold_left (fun l arg -> ltype_of_typ arg :: l) [] (List.rev rest)
        in
        L.pointer_type
          (L.function_type (ltype_of_typ rt) (Array.of_list llargs))
    | A.CompositeType (A.LIST, [ A.Int ]) -> L.pointer_type list_int_t
    | A.CompositeType (A.LIST, [ A.Bool ]) -> L.pointer_type list_bool_t
    | A.CompositeType (A.LIST, [ A.Float ]) -> L.pointer_type list_float_t
    | A.CompositeType (A.SET, [ _ ]) -> L.pointer_type set_t
    | A.CompositeType (A.DICT, [ _; _ ]) -> L.pointer_type dict_t
  in

  (* Create a map of global variables after creating each *)
  let global_vars : L.llvalue StringMap.t =
    let global_var m (t, n) =
      let init =
        match t with
        | A.Float -> L.const_float (ltype_of_typ t) 0.0
        | _ -> L.const_int (ltype_of_typ t) 0
      in
      StringMap.add n (L.define_global n init the_module) m
    in
    List.fold_left global_var StringMap.empty globals
  in

  let printf_t : L.lltype =
    L.var_arg_function_type i32_t [| L.pointer_type i8_t |]
  in
  let printf_func : L.llvalue =
    L.declare_function "printf" printf_t the_module
  in

  let printbig_t : L.lltype = L.function_type i32_t [| i32_t |] in
  let printbig_func : L.llvalue =
    L.declare_function "printbig" printbig_t the_module
  in

  let slice_t : L.lltype =
    L.function_type str_t [| L.pointer_type i8_t; i32_t; i32_t |]
  in
  let slice_func : L.llvalue = L.declare_function "slice" slice_t the_module in

  let length_t : L.lltype = L.function_type i32_t [| L.pointer_type i8_t |] in
  let length_func : L.llvalue =
    L.declare_function "length" length_t the_module
  in

  let concat_t : L.lltype =
    L.function_type str_t [| L.pointer_type i8_t; L.pointer_type i8_t |]
  in
  let concat_func : L.llvalue =
    L.declare_function "concat" concat_t the_module
  in

  let range_t : L.lltype =
    L.function_type (L.pointer_type list_int_t) [| i32_t |]
  in
  let range_func : L.llvalue = L.declare_function "range" range_t the_module in

  let create_empty_list_int_t : L.lltype =
    L.function_type (L.pointer_type list_int_t) [| i32_t |]
  in
  let create_empty_list_int_func : L.llvalue = L.declare_function "create_empty_list_int" create_empty_list_int_t the_module in
  
  let create_empty_list_bool_t : L.lltype =
    L.function_type (L.pointer_type list_bool_t) [| i32_t |]
  in
  let create_empty_list_bool_func : L.llvalue = L.declare_function "create_empty_list_bool" create_empty_list_bool_t the_module in
  
  let create_empty_list_float_t : L.lltype =
    L.function_type (L.pointer_type list_float_t) [| i32_t |]
  in
  let create_empty_list_float_func : L.llvalue = L.declare_function "create_empty_list_float" create_empty_list_float_t the_module in

  let len_t : L.lltype = L.function_type i32_t [| L.pointer_type i8_t |] in
  let len_func : L.llvalue = L.declare_function "len" len_t the_module in

  let append_t : L.lltype =
    L.function_type i32_t [| L.pointer_type i8_t; L.pointer_type i8_t |]
  in
  let append_func : L.llvalue =
    L.declare_function "append" append_t the_module
  in

  let idxtrans_t : L.lltype =
    L.function_type i32_t [| L.pointer_type i8_t; i32_t |]
  in
  let idxtrans_func : L.llvalue =
    L.declare_function "index_transform" idxtrans_t the_module
  in

  let list_slice_t : L.lltype =
    L.function_type
      (L.pointer_type list_int_t)
      [| L.pointer_type i8_t; i32_t; i32_t |]
  in
  let list_slice_func : L.llvalue =
    L.declare_function "list_slice" list_slice_t the_module
  in

  let create_empty_set_t : L.lltype =
    L.function_type (L.pointer_type set_t) [| i32_t |]
  in
  let create_empty_set_func : L.llvalue =
    L.declare_function "create_empty_set" create_empty_set_t the_module
  in

  let add_elem_t : L.lltype =
    L.function_type i32_t [| L.pointer_type set_t; L.pointer_type i8_t |]
  in
  let add_elem_func : L.llvalue =
    L.declare_function "add_elem" add_elem_t the_module
  in

  let search_elem_t : L.lltype =
    L.function_type i1_t [| L.pointer_type set_t; L.pointer_type i8_t |]
  in
  let search_elem_func : L.llvalue =
    L.declare_function "search_elem" search_elem_t the_module
  in

  let delete_elem_t : L.lltype =
    L.function_type i32_t [| L.pointer_type set_t; L.pointer_type i8_t |]
  in
  let delete_elem_func : L.llvalue =
    L.declare_function "delete_elem" delete_elem_t the_module
  in

  let get_set_size_t : L.lltype =
    L.function_type i32_t [| L.pointer_type set_t |]
  in
  let get_set_size_func : L.llvalue =
    L.declare_function "get_set_size" get_set_size_t the_module
  in

  let create_empty_dict_t : L.lltype =
    L.function_type (L.pointer_type dict_t) [| i32_t; i32_t |]
  in
  let create_empty_dict_func : L.llvalue =
    L.declare_function "create_empty_dict" create_empty_dict_t the_module
  in

  let add_key_value_t : L.lltype =
    L.function_type i32_t
      [| L.pointer_type dict_t; L.pointer_type i8_t; L.pointer_type i8_t |]
  in
  let add_key_value_func : L.llvalue =
    L.declare_function "add_key_value" add_key_value_t the_module
  in

  let get_dict_size_t : L.lltype =
    L.function_type i32_t [| L.pointer_type dict_t |]
  in
  let get_dict_size_func : L.llvalue =
    L.declare_function "get_dict_size" get_dict_size_t the_module
  in

  let key_exists_t : L.lltype =
    L.function_type i1_t [| L.pointer_type dict_t; L.pointer_type i8_t |]
  in
  let key_exists_func : L.llvalue =
    L.declare_function "key_exists" key_exists_t the_module
  in

  let retrieve_value_int_t : L.lltype =
    L.function_type i32_t [| L.pointer_type dict_t; L.pointer_type i8_t |]
  in
  let retrieve_value_int_func : L.llvalue =
    L.declare_function "retrieve_value_int" retrieve_value_int_t the_module
  in

  let retrieve_value_float_t : L.lltype =
    L.function_type float_t [| L.pointer_type dict_t; L.pointer_type i8_t |]
  in
  let retrieve_value_float_func : L.llvalue =
    L.declare_function "retrieve_value_float" retrieve_value_float_t the_module
  in

  let retrieve_value_bool_t : L.lltype =
    L.function_type i1_t [| L.pointer_type dict_t; L.pointer_type i8_t |]
  in
  let retrieve_value_bool_func : L.llvalue =
    L.declare_function "retrieve_value_bool" retrieve_value_bool_t the_module
  in

  let retrieve_value_list_t : L.lltype =
    L.function_type (L.pointer_type list_int_t) [| L.pointer_type dict_t; L.pointer_type i8_t |]
  in
  let retrieve_value_list_func : L.llvalue =
    L.declare_function "retrieve_value_list" retrieve_value_list_t the_module
  in

  let delete_key_t : L.lltype =
    L.function_type i32_t [| L.pointer_type dict_t; L.pointer_type i8_t |]
  in
  let delete_key_func : L.llvalue =
    L.declare_function "delete_key" delete_key_t the_module
  in

  (* Define each function (arguments and return type) so we can
     call it even before we've created its body *)
  let function_decls : (L.llvalue * sfunc_decl) StringMap.t =
    let function_decl m fdecl =
      let name = fdecl.sfname
      and formal_types =
        Array.of_list (List.map (fun (t, _) -> ltype_of_typ t) fdecl.sformals)
      in
      let ftype = L.function_type (ltype_of_typ fdecl.styp) formal_types in
      StringMap.add name (L.define_function name ftype the_module, fdecl) m
    in
    List.fold_left function_decl StringMap.empty functions
  in

  (* Fill in the body of the given function *)
  let rec build_function_body fdecl =
    let the_function, _ = StringMap.find fdecl.sfname function_decls in
    let builder = L.builder_at_end context (L.entry_block the_function) in

    let int_format_str = L.build_global_stringptr "%d\n" "fmt" builder
    and float_format_str = L.build_global_stringptr "%g\n" "fmt" builder
    and string_format_str = L.build_global_stringptr "%s\n" "fmt" builder in

    (* Construct the function's "locals": formal arguments and locally
       declared variables.  Allocate each on the stack, initialize their
       value, if appropriate, and remember their values in the "locals" map *)
    let local_vars =
      let local_vars_tbl = Hashtbl.create 10 in
      let add_formal (t, n) p =
        L.set_value_name n p;
        let local = L.build_alloca (ltype_of_typ t) n builder in
        ignore (L.build_store p local builder);
        Hashtbl.add local_vars_tbl n local
      in

      List.iter2 add_formal fdecl.sformals
        (Array.to_list (L.params the_function));
      local_vars_tbl
    in

    (* Return the value for a variable or formal argument.
       Check local names first, then global names *)
    let lookup n =
      try Hashtbl.find local_vars n
      with Not_found -> StringMap.find n global_vars
    in

    let list_type_to_idx typ =
      match typ with
      | A.CompositeType (A.LIST, [ A.Int ]) -> 0
      | A.CompositeType (A.LIST, [ A.Bool ]) -> 1
      | A.CompositeType (A.LIST, [ A.Float ]) -> 2
      (* | A.CompositeType(A.LIST, [A.String]) -> 3 *)
      | _ -> raise (Failure "Invalid list type")
    in
    let list_prim_type typ =
      match typ with
      | A.CompositeType (A.LIST, [ t ]) -> t
      | _ -> raise (Failure "Invalid list type")
    in

    let type_to_idx_map typ =
      match typ with
      | A.Int -> 0
      | A.Bool -> 1
      | A.Float -> 2
      | A.CompositeType(LIST, _) -> 4
      | _ ->
          raise (Failure "Internal error. Prime type has to be int/bool/float")
    in

    (* Construct code for an expression; return its value *)
    let rec expr builder ((et, e) : sexpr) =
      let alloc_var_and_return_voidptr_from_expr e =
        let e' = expr builder e in
        let p_e' = L.build_alloca (L.type_of e') "ptmp" builder in
        ignore (L.build_store e' p_e' builder);
        L.build_bitcast p_e' (L.pointer_type i8_t) "tmpvoidptr" builder
      in

      match e with
      | SLiteral i -> L.const_int i32_t i
      | SBoolLit b -> L.const_int i1_t (if b then 1 else 0)
      | SFliteral l -> L.const_float_of_string float_t l
      | SStrLit l -> L.build_global_stringptr l "string" builder
      (*@ *)
      (* | SFunctionLit s -> L.build_load (lookup s) s builder *)
      | SNoexpr -> L.const_int i32_t 0
      | SId s -> L.build_load (lookup s) s builder
      | SFunctionLit sfunc ->
          build_function_body sfunc;
          let fdef, fdecl = StringMap.find sfunc.sfname function_decls in
          fdef
      | SListExpr el ->
          let vl = List.map (expr builder) el in
          let e_typ = L.pointer_type (L.type_of (List.hd vl)) in
          let size = L.const_int i32_t (List.length vl) in
          let prim_type = list_prim_type et in
          let list_typ =
            match prim_type with
            | A.Int -> list_int_t
            | A.Bool -> list_bool_t
            | A.Float -> list_float_t
          in
          let typ_idx = list_type_to_idx et in
          (* Declare LIST *)
          let list = L.build_array_malloc e_typ size "list" builder in
          let list = L.build_pointercast list e_typ "list" builder in
          let list_struct = L.build_malloc list_typ "list_struct" builder in
          (* variable storing current size of list *)
          let list_size =
            L.build_struct_gep list_struct 0 "list_size" builder
          in
          ignore
            (L.build_store
               (L.const_int i32_t (List.length vl))
               list_size builder);
          (* variable storing type of list *)
          let list_type =
            L.build_struct_gep list_struct 1 "list_type" builder
          in
          ignore (L.build_store (L.const_int i32_t typ_idx) list_type builder);
          (* variable storing capacity of list, used later by append() *)
          let list_capacity =
            L.build_struct_gep list_struct 2 "list_capacity" builder
          in
          ignore
            (L.build_store
               (L.const_int i32_t (List.length vl))
               list_capacity builder);
          (* assign actual list with values *)
          let assign i v =
            let vp =
              L.build_gep list [| L.const_int i32_t i |] "list_v" builder
            in
            ignore (L.build_store v vp builder)
          in
          List.iteri assign vl;
          (* pointer to the actual list *)
          let list_ptr =
            L.build_struct_gep list_struct 3 "list_struct" builder
          in
          ignore (L.build_store list list_ptr builder);
          list_struct
      | SIndexAssign (s, e1, e2) ->
          let idx = expr builder e1
          and value = expr builder e2
          and list = L.build_load (lookup s) "getIndex" builder in
          (* Cast list to void * to send to len function *)
          let p_list = L.build_alloca (L.type_of list) "pcst" builder in
          ignore (L.build_store list p_list builder);
          let p_list =
            L.build_bitcast p_list (L.pointer_type i8_t) "cs" builder
          in
          let idx =
            L.build_call idxtrans_func [| p_list; idx |] "index_transform_call"
              builder
          in
          (* Get list pointer from struct *)
          let list = L.build_struct_gep list 3 "list_v" builder in
          let list = L.build_load list "list_v" builder in
          (* list[idx]*)
          let list_idx = L.build_gep list [| idx |] "getIndex_e" builder in
          ignore (L.build_store value list_idx builder);
          value
      | SAssign (s, e) ->
          let e' = expr builder e in
          ignore (L.build_store e' (lookup s) builder);
          e'
      | SGetIndex (s, e) ->
          let idx = expr builder e
          and list = L.build_load (lookup s) "getIndex" builder in
          let p_list = L.build_alloca (L.type_of list) "pcst" builder in
          ignore (L.build_store list p_list builder);
          let p_list =
            L.build_bitcast p_list (L.pointer_type i8_t) "cs" builder
          in
          let idx =
            L.build_call idxtrans_func [| p_list; idx |] "index_transform_call"
              builder
          in
          (* Get list pointer from struct *)
          let list = L.build_struct_gep list 3 "list_v" builder in
          let list = L.build_load list "list_v" builder in
          let sp = L.build_gep list [| idx |] "getIndex_e" builder in
          L.build_load sp "getIndex" builder
      | SGetSlice (s, e1, e2) ->
          let l = expr builder e1
          and r = expr builder e2
          and list = L.build_load (lookup s) "getlist" builder in
          (* Cast list to void * to send to len function *)
          let p_list = L.build_alloca (L.type_of list) "pcst" builder in
          ignore (L.build_store list p_list builder);
          let p_list =
            L.build_bitcast p_list (L.pointer_type i8_t) "cs" builder
          in
          (* Get list pointer from struct *)
          L.build_call list_slice_func [| p_list; l; r |] "list_slice_call"
            builder
      | SBinop (((A.Float, _) as e1), op, e2) ->
          let e1' = expr builder e1 and e2' = expr builder e2 in
          (match op with
          | A.Add -> L.build_fadd
          | A.Sub -> L.build_fsub
          | A.Mult -> L.build_fmul
          | A.Div -> L.build_fdiv
          | A.Equal -> L.build_fcmp L.Fcmp.Oeq
          | A.Neq -> L.build_fcmp L.Fcmp.One
          | A.Less -> L.build_fcmp L.Fcmp.Olt
          | A.Leq -> L.build_fcmp L.Fcmp.Ole
          | A.Greater -> L.build_fcmp L.Fcmp.Ogt
          | A.Geq -> L.build_fcmp L.Fcmp.Oge
          | A.And | A.Or ->
              raise
                (Failure
                   "internal error: semant should have rejected and/or on float"))
            e1' e2' "tmp" builder
      | SBinop (e1, op, e2) ->
          let e1' = expr builder e1 and e2' = expr builder e2 in
          (match op with
          | A.Add -> L.build_add
          | A.Sub -> L.build_sub
          | A.Mult -> L.build_mul
          | A.Div -> L.build_sdiv
          | A.And -> L.build_and
          | A.Or -> L.build_or
          | A.Equal -> L.build_icmp L.Icmp.Eq
          | A.Neq -> L.build_icmp L.Icmp.Ne
          | A.Less -> L.build_icmp L.Icmp.Slt
          | A.Leq -> L.build_icmp L.Icmp.Sle
          | A.Greater -> L.build_icmp L.Icmp.Sgt
          | A.Geq -> L.build_icmp L.Icmp.Sge)
            e1' e2' "tmp" builder
      | SUnop (op, ((t, _) as e)) ->
          let e' = expr builder e in
          (match op with
          | A.Neg when t = A.Float -> L.build_fneg
          | A.Neg -> L.build_neg
          | A.Not -> L.build_not)
            e' "tmp" builder
      | SCall ("print", [ e ]) | SCall ("printb", [ e ]) ->
          L.build_call printf_func
            [| int_format_str; expr builder e |]
            "printf" builder
      | SCall ("printbig", [ e ]) ->
          L.build_call printbig_func [| expr builder e |] "printbig" builder
      | SCall ("printf", [ e ]) ->
          L.build_call printf_func
            [| float_format_str; expr builder e |]
            "printf" builder
      | SCall ("prints", [ e ]) ->
          L.build_call printf_func
            [| string_format_str; expr builder e |]
            "printf" builder
      | SCall ("slice", [ e; e1; e2 ]) ->
          L.build_call slice_func
            [| expr builder e; expr builder e1; expr builder e2 |]
            "slice" builder
      | SCall ("length", [ e ]) ->
          L.build_call length_func [| expr builder e |] "length" builder
      | SCall ("concat", [ e; e1 ]) ->
          L.build_call concat_func
            [| expr builder e; expr builder e1 |]
            "concat" builder
      | SCall ("range", [ e ]) ->
          L.build_call range_func [| expr builder e |] "range_call" builder
      | SCall ("len", [ e ]) ->
          let e' = expr builder e in
          (* Cast list to void * to send to len function *)
          let p_e' = L.build_alloca (L.type_of e') "pcst" builder in
          ignore (L.build_store e' p_e' builder);
          let e' = L.build_bitcast p_e' (L.pointer_type i8_t) "cs" builder in
          L.build_call len_func [| e' |] "list_len_call" builder
      | SCall ("append", [ e1; e2 ]) ->
          let e1' = expr builder e1 in
          let e2' = expr builder e2 in
          (* Cast list to void * to send to append function *)
          let p_e1' = L.build_alloca (L.type_of e1') "pcst" builder in
          ignore (L.build_store e1' p_e1' builder);
          let p_e2' = L.build_alloca (L.type_of e2') "pcst" builder in
          ignore (L.build_store e2' p_e2' builder);
          let e1' = L.build_bitcast p_e1' (L.pointer_type i8_t) "cs" builder in
          let e2' = L.build_bitcast p_e2' (L.pointer_type i8_t) "cs" builder in
          L.build_call append_func [| e1'; e2' |] "list_append_call" builder
      | SCall ("setAdd", [ se; e ]) ->
          let e_voidptr = alloc_var_and_return_voidptr_from_expr e in
          L.build_call add_elem_func
            [| expr builder se; e_voidptr |]
            "add_elem" builder
      | SCall ("setFind", [ se; e ]) ->
          let e_voidptr = alloc_var_and_return_voidptr_from_expr e in
          L.build_call search_elem_func
            [| expr builder se; e_voidptr |]
            "search_elem" builder
      | SCall ("setRemove", [ se; e ]) ->
          let e_voidptr = alloc_var_and_return_voidptr_from_expr e in
          L.build_call delete_elem_func
            [| expr builder se; e_voidptr |]
            "delete_elem" builder
      | SCall ("setSize", [ se ]) ->
          L.build_call get_set_size_func
            [| expr builder se |]
            "get_set_size" builder
      | SCall ("dictAdd", [ de; e1; e2 ]) ->
          let e1_voidptr = alloc_var_and_return_voidptr_from_expr e1 in
          let e2_voidptr = alloc_var_and_return_voidptr_from_expr e2 in
          L.build_call add_key_value_func
            [| expr builder de; e1_voidptr; e2_voidptr |]
            "add_key_value" builder
      | SCall ("dictSize", [ de ]) ->
          L.build_call get_dict_size_func
            [| expr builder de |]
            "get_dict_size" builder
      | SCall ("dictHasKey", [ de; e ]) ->
          let e_voidptr = alloc_var_and_return_voidptr_from_expr e in
          L.build_call key_exists_func
            [| expr builder de; e_voidptr |]
            "key_exists" builder
      | SCall ("dictGetInt", [ de; e ]) ->
          let e_voidptr = alloc_var_and_return_voidptr_from_expr e in
          L.build_call retrieve_value_int_func
            [| expr builder de; e_voidptr |]
            "retrieve_value_int" builder
      | SCall ("dictGetBool", [ de; e ]) ->
          let e_voidptr = alloc_var_and_return_voidptr_from_expr e in
          L.build_call retrieve_value_bool_func
            [| expr builder de; e_voidptr |]
            "retrieve_value_bool" builder
      | SCall ("dictGetFloat", [ de; e ]) ->
          let e_voidptr = alloc_var_and_return_voidptr_from_expr e in
          L.build_call retrieve_value_float_func
            [| expr builder de; e_voidptr |]
            "retrieve_value_float" builder
      | SCall ("dictGetList", [ de; e ]) ->
          let e_voidptr = alloc_var_and_return_voidptr_from_expr e in
          L.build_call retrieve_value_list_func
            [| expr builder de; e_voidptr |]
            "retrieve_value_list" builder
      | SCall ("dictRemove", [ de; e ]) ->
          let e_voidptr = alloc_var_and_return_voidptr_from_expr e in
          L.build_call delete_key_func
            [| expr builder de; e_voidptr |]
            "delete_key" builder
      | SCall (f, args) ->
          let result = match et with A.Void -> "" | _ -> f ^ "_result" in
          let llargs = List.rev (List.map (expr builder) (List.rev args)) in
          let fdef =
            if StringMap.mem f function_decls then
              fst (StringMap.find f function_decls)
            else L.build_load (lookup f) f builder
          in
          L.build_call fdef (Array.of_list llargs) result builder
    in

    (* LLVM insists each basic block end with exactly one "terminator"
       instruction that transfers control.  This function runs "instr builder"
       if the current block does not already have a terminator.  Used,
       e.g., to handle the "fall off the end of the function" case. *)
    let add_terminal builder instr =
      match L.block_terminator (L.insertion_block builder) with
      | Some _ -> ()
      | None -> ignore (instr builder)
    in

    (* Build the code for the given statement; return the builder for
       the statement's successor (i.e., the next instruction will be built
       after the one generated by this call) *)
    let rec stmt builder = function
      | SBlock sl -> List.fold_left stmt builder sl
      | SExpr e ->
          ignore (expr builder e);
          builder
      | SReturn e ->
          ignore
            (match fdecl.styp with
            (* Special "return nothing" instr *)
            | A.Void -> L.build_ret_void builder (* Build return statement *)
            | _ -> L.build_ret (expr builder e) builder);
          builder
      | SIf (predicate, then_stmt, else_stmt) ->
          let bool_val = expr builder predicate in
          let merge_bb = L.append_block context "merge" the_function in
          let build_br_merge = L.build_br merge_bb in

          (* partial function *)
          let then_bb = L.append_block context "then" the_function in
          add_terminal
            (stmt (L.builder_at_end context then_bb) then_stmt)
            build_br_merge;

          let else_bb = L.append_block context "else" the_function in
          add_terminal
            (stmt (L.builder_at_end context else_bb) else_stmt)
            build_br_merge;

          ignore (L.build_cond_br bool_val then_bb else_bb builder);
          L.builder_at_end context merge_bb
      | SWhile (predicate, body) ->
          let pred_bb = L.append_block context "while" the_function in
          ignore (L.build_br pred_bb builder);

          let body_bb = L.append_block context "while_body" the_function in
          add_terminal
            (stmt (L.builder_at_end context body_bb) body)
            (L.build_br pred_bb);

          let pred_builder = L.builder_at_end context pred_bb in
          let bool_val = expr pred_builder predicate in

          let merge_bb = L.append_block context "merge" the_function in
          ignore (L.build_cond_br bool_val body_bb merge_bb pred_builder);
          L.builder_at_end context merge_bb
      (* Implement for loops as while loops *)
      | SFor (e1, e2, e3, body) ->
          stmt builder
            (SBlock [ SExpr e1; SWhile (e2, SBlock [ body; SExpr e3 ]) ])
      | SForin (iter, clist, x, pred, inc, next, body) ->
          stmt builder
            (SBlock
               [
                 iter;
                 clist;
                 x;
                 SWhile (pred, SBlock [ body; SExpr inc; SExpr next ]);
               ])
      | SVarDecl (t, id) ->
          let local_var = L.build_alloca (ltype_of_typ t) id builder in
          Hashtbl.add local_vars id local_var;
          ignore
            (match t with
            | CompositeType (LIST, [ Int ]) ->
                ignore
                  (L.build_store
                     (L.build_call create_empty_list_int_func
                        [| L.const_int i32_t (type_to_idx_map Int) |]
                        "create_empty_list_int" builder)
                     local_var builder)
            | CompositeType (LIST, [ Bool ]) ->
                ignore
                  (L.build_store
                     (L.build_call create_empty_list_bool_func
                        [| L.const_int i32_t (type_to_idx_map Bool) |]
                        "create_empty_list_bool" builder)
                     local_var builder)
            | CompositeType (LIST, [ Float ]) ->
                ignore
                  (L.build_store
                     (L.build_call create_empty_list_float_func
                        [| L.const_int i32_t (type_to_idx_map Float) |]
                        "create_empty_list_float" builder)
                     local_var builder)
            | CompositeType (SET, [ prim_typ ]) ->
                ignore
                  (L.build_store
                     (L.build_call create_empty_set_func
                        [| L.const_int i32_t (type_to_idx_map prim_typ) |]
                        "create_empty_set" builder)
                     local_var builder)
            | CompositeType (DICT, [ key_typ; value_typ ]) ->
                ignore
                  (L.build_store
                     (L.build_call create_empty_dict_func
                        [|
                          L.const_int i32_t (type_to_idx_map key_typ);
                          L.const_int i32_t (type_to_idx_map value_typ);
                        |]
                        "create_empty_dict" builder)
                     local_var builder)
            | _ -> ());
          builder
      | SVarDeclAssign ((t, id), e) ->
          let local_var = L.build_alloca (ltype_of_typ t) id builder in
          Hashtbl.add local_vars id local_var;
          let e' = expr builder e in
          ignore (L.build_store e' (lookup id) builder);
          builder
    in

    (* Build the code for each statement in the function *)
    let builder = stmt builder (SBlock fdecl.sbody) in

    (* Add a return if the last block falls off the end *)
    add_terminal builder
      (match fdecl.styp with
      | A.Void -> L.build_ret_void
      | A.Float -> L.build_ret (L.const_float float_t 0.0)
      | t -> L.build_ret (L.const_int (ltype_of_typ t) 0))
  in

  List.iter build_function_body top_funcs;
  the_module
