(* Semantic checking for the FFBB compiler *)

open Ast
open Sast
module StringMap = Map.Make (String)

(* Semantic checking of the AST. Returns an SAST if successful,
   throws an exception if something is wrong.

   Check each global variable, then check each function *)
let local_sfunc = ref []

let function_decls = ref StringMap.empty

let check (globals, functions) =
  (* Verify a list of bindings has no void types or duplicate names *)
  let check_binds (kind : string) (binds : bind list) =
    List.iter
      (function
        | Void, b -> raise (Failure ("illegal void " ^ kind ^ " " ^ b))
        | _ -> ())
      binds;
    let rec dups = function
      | [] -> ()
      | (_, n1) :: (_, n2) :: _ when n1 = n2 ->
          raise (Failure ("duplicate " ^ kind ^ " " ^ n1))
      | _ :: t -> dups t
    in
    dups (List.sort (fun (_, a) (_, b) -> compare a b) binds)
  in

  (* Verify the new binding to be added is not a duplicate *)
  let check_dup (binding : bind) (bindings : bindtbl) =
    match binding with
    | Void, s -> raise (Failure ("illegal void local " ^ s))
    | _, s ->
        if Hashtbl.mem bindings s then raise (Failure ("duplicate local " ^ s))
        else ()
  in

  (**** Check global variables ****)
  check_binds "global" globals;

  (**** Check functions ****)
  let lambda_num = ref 0 in

  (* Collect function declarations for built-in functions: no bodies *)
  let get_atypes fdecl = List.map (fun x -> fst x) fdecl.formals in
  let get_ftype fdecl = FunctionType (fdecl.typ :: get_atypes fdecl) in
  let built_in_decls =
    let add_bind map (name, return_ty, tys) =
      StringMap.add name
        {
          typ = return_ty;
          fname = name;
          formals =
            (match List.length tys with
            | 0 -> []
            | 1 -> [ (List.nth tys 0, "arg1") ]
            | 2 -> [ (List.nth tys 0, "arg1"); (List.nth tys 1, "arg2") ]
            | 3 ->
                [
                  (List.nth tys 0, "arg1");
                  (List.nth tys 1, "arg2");
                  (List.nth tys 2, "arg3");
                ]
            | _ -> raise (Failure "Built-in function must take 0-3 arguments"));
          body = [];
        }
        map
    in
    List.fold_left add_bind StringMap.empty
      [
        ("print", Void, [ Int ]);
        ("printb", Void, [ Bool ]);
        ("prints", Void, [ String ]);
        ("printf", Void, [ Float ]);
        ("printbig", Void, [ Int ]);
        ("setAdd", Void, [ CompositeType (SET, [ Int ]); Int ]);
        ("setFind", Bool, [ CompositeType (SET, [ Int ]); Int ]);
        ("setRemove", Void, [ CompositeType (SET, [ Int ]); Int ]);
        ("setSize", Int, [ CompositeType (SET, [ Int ]) ]);
        ("dictAdd", Void, [ CompositeType (DICT, [ Int; Int ]); Int; Int ]);
        ("dictSize", Int, [ CompositeType (DICT, [ Int; Int ]) ]);
        ("dictHasKey", Bool, [ CompositeType (DICT, [ Int; Int ]); Int ]);
        ("dictGetInt", Int, [ CompositeType (DICT, [ Int; Int ]); Int ]);
        ("dictGetBool", Bool, [ CompositeType (DICT, [ Int; Int ]); Int ]);
        ("dictGetFloat", Float, [ CompositeType (DICT, [ Int; Int ]); Int ]);
        ("dictGetList", CompositeType (LIST, [Int]), [ CompositeType (DICT, [ Int; Int ]); Int ]);
        ("dictRemove", Int, [ CompositeType (DICT, [ Int; Int ]); Int ]);
        ("len", Int, [ CompositeType (LIST, [ Void ]) ]);
        ("range", CompositeType (LIST, [ Int ]), [ Int ]);
        ("append", Void, [ CompositeType (LIST, [ Void ]); Void ]);
        ("slice", String, [ String; Int; Int ]);
        ("length", Int, [ String ]);
        ("concat", String, [ String; String ]);
      ]
  in

  (* Add function name to symbol table *)
  let add_func map fd =
    let built_in_err = "function " ^ fd.fname ^ " may not be defined"
    and dup_err = "duplicate function " ^ fd.fname
    and make_err er = raise (Failure er)
    and n = fd.fname (* Name of the function *) in
    match fd with
    (* No duplicate functions or redefinitions of built-ins *)
    | _ when StringMap.mem n built_in_decls -> make_err built_in_err
    | _ when StringMap.mem n map -> make_err dup_err
    | _ -> StringMap.add n fd map
  in

  (* Collect all function names into one symbol table *)
  let _ =
    ignore (function_decls := List.fold_left add_func built_in_decls functions)
  in

  (* Return a function from our symbol table *)
  let find_func s =
    try StringMap.find s !function_decls
    with Not_found -> raise (Failure ("unrecognized function " ^ s))
  in

  let _ = find_func "main" in

  (* Ensure "main" is defined *)
  let rec check_function func =
    let bindings = Hashtbl.create 10 in
    (* Make sure no formals are void or duplicates *)
    check_binds "formal" func.formals;

    let add_formal_to_bindings formal =
      Hashtbl.add bindings (snd formal) (fst formal)
    in
    List.iter add_formal_to_bindings func.formals;

    (* Raise an exception if the given rvalue type cannot be assigned to
       the given lvalue type *)
    let check_assign fname lvaluet rvaluet err =
      match fname with
      | "append" | "len" | "setAdd" | "setFind" | "setRemove" | "setSize"
      | "dictAdd" | "dictHasKey" | "dictGetInt" | "dictGetBool" | "dictGetFloat" | "dictGetList"
      | "dictRemove" | "dictSize" ->
          lvaluet
      | _ -> if lvaluet = rvaluet then lvaluet else raise (Failure err)
    in

    (* Raise an exception if the given rvalue type cannot be assigned to
       the given lvalue type *)
    let get_list_prim_typ t =
      match t with CompositeType (LIST, [ typ ]) -> typ
    in
    let check_get_index lvaluet rvaluet err = get_list_prim_typ lvaluet in

    (* Build local symbol table of variables for this function *)
    let symbols =
      List.fold_left
        (fun m (ty, name) -> StringMap.add name ty m)
        StringMap.empty globals
    in

    (* Return a variable from our local symbol table *)
    let type_of_identifier s =
      try Hashtbl.find bindings s
      with Not_found -> (
        try StringMap.find s symbols
        with Not_found -> raise (Failure ("undeclared identifier " ^ s)))
    in

    (* Return a semantically-checked expression, i.e., with a type *)
    let rec expr = function
      | Literal l -> (Int, SLiteral l)
      | Fliteral l -> (Float, SFliteral l)
      | BoolLit l -> (Bool, SBoolLit l)
      | FunctionLit l ->
          function_decls := List.fold_left add_func !function_decls [ l ];
          let local_sf = check_function l in
          let _ = local_sfunc := local_sf :: !local_sfunc in
          (get_ftype l, SFunctionLit local_sf)
      | StrLit l -> (String, SStrLit l)
      | Noexpr -> (Void, SNoexpr)
      | Id s -> (type_of_identifier s, SId s)
      | IndexAssign (var, e1, e2) as ex ->
          let lt = type_of_identifier var
          and rt1, e1' = expr e1
          and rt2, e2' = expr e2 in
          let err = "illegal index assignment" in
          (lt, SIndexAssign (var, (rt1, e1'), (rt2, e2')))
      | GetIndex (var, e2) as ex ->
          let lt = type_of_identifier var and rt, e' = expr e2 in
          let err =
            "illegal index access " ^ string_of_typ lt ^ "[" ^ string_of_expr ex
            ^ "]"
          in
          (check_get_index lt rt err, SGetIndex (var, (rt, e')))
      | GetSlice (var, e1, e2) as ex ->
          let lt = type_of_identifier var
          and rt1, e1' = expr e1
          and rt2, e2' = expr e2 in
          let err =
            "illegal slice access " ^ string_of_typ lt ^ "[" ^ string_of_expr ex
            ^ "]"
          in
          (lt, SGetSlice (var, (rt1, e1'), (rt2, e2')))
      | Assign (var, e) as ex ->
          let lt = type_of_identifier var and rt, e' = expr e in
          let err =
            "illegal assignment " ^ string_of_typ lt ^ " = " ^ string_of_typ rt
            ^ " in " ^ string_of_expr ex
          in
          (check_assign var lt rt err, SAssign (var, (rt, e')))
      | ListExpr el as ex ->
          let check_list e =
            let et, e' = expr e in
            (et, e')
          in
          let el' = List.map check_list el in
          let head = List.hd el' in
          let get_rt (a, _) = [ a ] in
          (CompositeType (LIST, get_rt head), SListExpr el')
      | Unop (op, e) as ex -> (
          let t, e' = expr e in
          let ty =
            match op with
            | Neg when t = Int || t = Float -> t
            | Not when t = Bool -> Bool
            | (Inc | Dec) when t = Int -> t
            | Inc | Dec ->
                raise
                  (Failure
                     ("illegal unary operator " ^ string_of_typ t
                    ^ string_of_uop op ^ " in " ^ string_of_expr ex))
            | _ ->
                raise
                  (Failure
                     ("illegal unary operator " ^ string_of_uop op
                    ^ string_of_typ t ^ " in " ^ string_of_expr ex))
          in
          match ex with
          | Unop (Neg, _) | Unop (Not, _) -> (ty, SUnop (op, (t, e')))
          | Unop (Inc, Id s) ->
              (ty, SAssign (s, (t, SBinop ((t, SId s), Add, (t, SLiteral 1)))))
          | Unop (Dec, Id s) ->
              (ty, SAssign (s, (t, SBinop ((t, SId s), Sub, (t, SLiteral 1))))))
      | IndexUnop (v, e, o) as ex -> (
          let list_t = type_of_identifier v and index_t, e' = expr e in
          let err =
            "illegal index access " ^ string_of_typ list_t ^ "["
            ^ string_of_expr ex ^ "]"
          in
          let index_element =
            (check_get_index list_t index_t err, SGetIndex (v, (index_t, e')))
          in
          match o with
          | Inc ->
              ( list_t,
                SIndexAssign
                  ( v,
                    (index_t, e'),
                    (Int, SBinop (index_element, Add, (Int, SLiteral 1))) ) )
          | Dec ->
              ( list_t,
                SIndexAssign
                  ( v,
                    (index_t, e'),
                    (Int, SBinop (index_element, Sub, (Int, SLiteral 1))) ) )
          | _ -> raise (Failure "Case should never be reached"))
      | Binop (e1, op, e2) as e ->
          let t1, e1' = expr e1 and t2, e2' = expr e2 in
          (* All binary operators require operands of the same type *)
          let same = t1 = t2 in
          (* Determine expression type based on operator and operand types *)
          let ty =
            match op with
            | (Add | Sub | Mult | Div) when same && t1 = Int -> Int
            | (Add | Sub | Mult | Div) when same && t1 = Float -> Float
            | (Equal | Neq) when same -> Bool
            | (Less | Leq | Greater | Geq) when same && (t1 = Int || t1 = Float)
              ->
                Bool
            | (And | Or) when same && t1 = Bool -> Bool
            | _ ->
                raise
                  (Failure
                     ("illegal binary operator " ^ string_of_typ t1 ^ " "
                    ^ string_of_op op ^ " " ^ string_of_typ t2 ^ " in "
                    ^ string_of_expr e))
          in
          (ty, SBinop ((t1, e1'), op, (t2, e2')))
      | Call (fname, args) as call ->
          let fd =
            if StringMap.mem fname !function_decls then find_func fname
            else if Hashtbl.mem bindings fname then
              match type_of_identifier fname with
              | FunctionType (rt :: rest) ->
                  {
                    typ = rt;
                    fname;
                    formals = List.mapi (fun i x -> (x, string_of_int i)) rest;
                    body = [];
                  }
              | _ -> raise (Failure ("unrecognized function " ^ fname))
            else raise (Failure ("unrecognized function " ^ fname))
          in
          let param_length = List.length fd.formals in
          if List.length args != param_length then
            raise
              (Failure
                 ("expecting " ^ string_of_int param_length ^ " arguments in "
                ^ string_of_expr call))
          else
            (* check if formal_type == expr_type *)
            let _check_call fname ft e =
              let et, e' = expr e in
              let err =
                "illegal argument found " ^ string_of_typ et ^ " expected "
                ^ string_of_typ ft ^ " in " ^ string_of_expr e
              in
              (check_assign fname ft et err, e')
            in
            let get_et e =
              let et, e' = expr e in
              et
            in
            let err_msg = fname ^ " type incorrect" in
            let err_msg_2 =
              "first argument of " ^ fname ^ " should have composite type"
            in
            (* set arg_idx to -1 when internal type check is no needed *)
            let check_custom_fun_type exp_high_typ len arg_idxs =
              match List.map get_et args with
              | CompositeType (high_typ, [ prim_typ ]) :: rest ->
                  if
                    List.length rest = len
                    && high_typ = exp_high_typ
                    &&
                    match arg_idxs with
                    | [ arg_idx ] -> prim_typ = List.nth rest arg_idx
                    | [] -> true
                    | _ -> raise (Failure "Internal error at arg_idxs")
                  then fname
                  else raise (Failure err_msg)
              | CompositeType (high_typ, [ key_typ; value_typ ]) :: rest ->
                  if
                    List.length rest = len
                    && high_typ = exp_high_typ
                    &&
                    match arg_idxs with
                    | [ arg_idx1; arg_idx2 ] ->
                        key_typ = List.nth rest arg_idx1
                        && value_typ = List.nth rest arg_idx2
                    | [ arg_idx ] -> key_typ = List.nth rest arg_idx
                    | [] -> true
                    | _ -> raise (Failure "Internal error at arg_idxs")
                  then fname
                  else raise (Failure err_msg)
              | _ -> raise (Failure err_msg_2)
            in
            let _ =
              match fname with
              | "append" -> check_custom_fun_type LIST 1 [ 0 ]
              | "len" -> check_custom_fun_type LIST 0 []
              | "setAdd" -> check_custom_fun_type SET 1 [ 0 ]
              | "setFind" -> check_custom_fun_type SET 1 [ 0 ]
              | "setRemove" -> check_custom_fun_type SET 1 [ 0 ]
              | "setSize" -> check_custom_fun_type SET 0 []
              | "dictAdd" -> check_custom_fun_type DICT 2 [ 0; 1 ]
              | "dictHasKey" -> check_custom_fun_type DICT 1 [ 0 ]
              | "dictGetInt" -> check_custom_fun_type DICT 1 [ 0 ]
              | "dictGetBool" -> check_custom_fun_type DICT 1 [ 0 ]
              | "dictGetFloat" -> check_custom_fun_type DICT 1 [ 0 ]
              | "dictGetList" -> check_custom_fun_type DICT 1 [ 0 ]
              | "dictRemove" -> check_custom_fun_type DICT 1 [ 0 ]
              | "dictSize" -> check_custom_fun_type DICT 0 []
              | _ -> fname
            in
            let check_call (ft, _) e = _check_call fname ft e in
            let args' = List.map2 check_call fd.formals args in
            (fd.typ, SCall (fname, args'))
    in

    let check_bool_expr e =
      let t', e' = expr e
      and err = "expected Boolean expression in " ^ string_of_expr e in
      if t' != Bool then raise (Failure err) else (t', e')
    in

    (* Return a semantically-checked statement i.e. containing sexprs *)
    let rec check_stmt = function
      | Expr e -> SExpr (expr e)
      | If (p, b1, b2) -> SIf (check_bool_expr p, check_stmt b1, check_stmt b2)
      | For (e1, e2, e3, st) ->
          SFor (expr e1, check_bool_expr e2, expr e3, check_stmt st)
      | Forin (e1, e2, st) ->
          lambda_num := !lambda_num + 1;
          let rt, e2' = expr e2 in
          let lt = get_list_prim_typ rt in
          let iter_n = "iter" ^ string_of_int !lambda_num in
          let clist_n = "clist" ^ string_of_int !lambda_num in
          let iter = check_stmt (VarDeclAssign ((Int, iter_n), Literal 0)) in
          let clist = check_stmt (VarDeclAssign ((rt, clist_n), e2)) in
          let pred =
            expr (Binop (Id iter_n, Less, Call ("len", [ Id clist_n ])))
          in
          let inc = expr (Unop (Inc, Id iter_n)) in
          let x =
            check_stmt (VarDeclAssign ((lt, e1), GetIndex (clist_n, Id iter_n)))
          in
          let next = expr (Assign (e1, GetIndex (clist_n, Id iter_n))) in
          SForin (iter, clist, x, pred, inc, next, check_stmt st)
      | While (p, s) -> SWhile (check_bool_expr p, check_stmt s)
      | Return e ->
          let t, e' = expr e in
          if t = func.typ then SReturn (t, e')
          else
            raise
              (Failure
                 ("return gives " ^ string_of_typ t ^ " expected "
                ^ string_of_typ func.typ ^ " in " ^ string_of_expr e))
      (* A block is correct if each statement is correct and nothing
         follows any Return statement.  Nested blocks are flattened. *)
      | Block sl ->
          let rec check_stmt_list = function
            | [ (Return _ as s) ] -> [ check_stmt s ]
            | Return _ :: _ -> raise (Failure "nothing may follow a return")
            | Block sl :: ss -> check_stmt_list (sl @ ss) (* Flatten blocks *)
            | s :: ss ->
                let tmp = check_stmt s in
                tmp :: check_stmt_list ss
            | [] -> []
          in
          SBlock (check_stmt_list sl)
      | VarDecl (t, s) ->
          check_dup (t, s) bindings;
          Hashtbl.add bindings s t;
          SVarDecl (t, s)
      | VarDeclAssign ((t, s), e) as stmt ->
          check_dup (t, s) bindings;
          Hashtbl.add bindings s t;
          let lt = t and rt, e' = expr e in
          let err =
            "illegal assignment " ^ string_of_typ lt ^ " = " ^ string_of_typ rt
            ^ " in " ^ string_of_stmt stmt
          in
          let tmp =
            (check_assign s lt rt err, SVarDeclAssign ((t, s), (rt, e')))
          in
          snd tmp
    in

    (* body of check_function *)
    {
      styp = func.typ;
      sfname = func.fname;
      sformals = func.formals;
      sftype = get_ftype func;
      slocals = bindings;
      sbody =
        (match check_stmt (Block func.body) with
        | SBlock sl -> sl
        | _ -> raise (Failure "internal error: block didn't become a block?"));
    }
  in
  let sfunc_decls = List.map check_function functions in
  ((globals, List.rev !local_sfunc @ sfunc_decls), sfunc_decls)
